#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <arpa/tftp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include "tftpaccess.h"

/* Enhancements to tftpd made at Yale:
 * 07-Mar-89
 * Alan S. Watt
 * watt-alan@cs.yale.edu
 *
 * Control such things as default directory and debugging level from
 * configuration file (/usr/local/etc/tftpd.conf)
 *
 * Permit access lists to say which systems are allowed access over
 * specific files.
 *
 * 12-Sept-89
 * Alan S. Watt
 * watt-alan@cs.yale.edu
 *
 * Modified the access list maintenance so exact matches (no mask bits)
 * are kept at the head of the list and permission searches end on the
 * first address match.  This has the effect of giving exact matches
 * a higher priority than wildcard matches.  This is also the way the
 * cisco access lists work.
 */

/*  We export only the following globals:
 *
 * char *tftpDefaultDirectory;
 * char *tftpRootDirectory;
 * int tftpDebugLevel;
 * int configLineNumber;
 * int addAccessList(int argc, char** argv);
 * int addFileRestriction(int argc, char** argv);
 * int validateAccessToFile(char* filename, unsigned long sourceAddress);
 * int readConfigFile(int argc, char** argv);
 * int realPath(char* orig, char* buf);
 *
 */


/* location of the configuration file */
int	configLineNumber;

/* If set, it means we have chdir() and chroot() to this directory and
 * unqualified pathnames are OK.  Otherwise they are illegal.
 */
char*	tftpDefaultDirectory	= (char*)0;
char*	tftpRootDirectory	= (char*)0;
int	tftpDebugLevel		= 0;
int	maxInputWait		= 5*60;		/* default wait for 5 minutes */

/* declare locals required for access lists
 */
static char*		accessFormatError;
static listhead_t*	accessList;
static char*		cfgets();
static struct stat	configStat;
static int		defaultAccessList;
static fileaccess_t*	fileHead[NFILEHEAD];
static int		maxAccessList;
static unsigned int	namehash();
static int		split();

/* for debugging printouts */
static char* permNames[] ={ "deny", "permit", "readonly", "readwrite" };
#define NITEMS(X)	(sizeof(X) / sizeof (X)[0])
#define PERMNAME(N)	((unsigned)(N) >= NITEMS(permNames) \
				? "<unknown>" : permNames[N])

#define CMD_DEFAULT_DIR		1
#define CMD_DEBUG_LEVEL		2
#define CMD_ROOT_DIR		3
#define CMD_ACCESS_LIST		4
#define CMD_DEFAULT_ACCESS_LIST	5
#define CMD_RESTRICT		6
#define CMD_INPUT_WAIT		7
static struct CMDS {
	char* cmdName;
	int cmdVal;
} configCmds[] ={
	"default-directory",	CMD_DEFAULT_DIR,
	"defaultDirectory",	CMD_DEFAULT_DIR,
	"debug-level",		CMD_DEBUG_LEVEL,
	"debugLevel",		CMD_DEBUG_LEVEL,
	"root-directory",	CMD_ROOT_DIR,
	"rootDirectory",	CMD_ROOT_DIR,
	"access-list",		CMD_ACCESS_LIST,
	"accessList",		CMD_ACCESS_LIST,
	"default-access-list",	CMD_DEFAULT_ACCESS_LIST,
	"defaultAccessList",	CMD_DEFAULT_ACCESS_LIST,
	"restrict",		CMD_RESTRICT,
	"inputWait",		CMD_INPUT_WAIT,
	"input-wait",		CMD_INPUT_WAIT,
	(char*)0,		0
};

/* Add an access list specification */
addAccessList (argc, argv)
int	argc;
char	**argv;
{
	condition_t condition;
	int list;

	if ((list = compileAccessCondition(argc, argv, &condition)) != 0)
		addConditionToList (&condition, list);
	else {
		syslog (LOG_DEBUG, "Line %d: access list format error = '%s'",
			configLineNumber, accessFormatError);
	}
}

/* Add an access condition to access list <list>.
 * An access list is maintained in the order specified, except
 * that exact matches are at the front of the list and wildcard
 * matches are at the end.
 */
static
addConditionToList (condition, list)
condition_t* condition;
int list;
{
	condition_t* cp, * lp;
	listhead_t* hp;

	/* get enough space for the condition */
	cp = (condition_t*)malloc(sizeof *condition);

	/* copy argument into allocated storage */
	*cp = *condition;

	/* See if we need to make global list bigger
	 * If so, reallocate it and clear any extra.
	 */
	if (list > maxAccessList) {
		int newsize = (list+1)*sizeof(accessList[0]);
		int oldsize = (maxAccessList+1)*sizeof(accessList[0]);
		listhead_t* newlist;

		/* get new space and clear it */
		newlist = (listhead_t*)malloc(newsize);
		bzero ((char*)newlist, newsize);

		/* if there used to be space, copy it */
		if (accessList)
			bcopy ((char*)accessList, (char*)newlist, oldsize);

		maxAccessList = list;
		accessList = newlist;
	}

	/* if this particular list has no conditions, put this one
	 * at the head.  Set 'lastExact' if this condition is not
	 * a wildcard.
	 */
	hp = &accessList[list];
	if ((lp = hp->tail) == 0) {
		hp->tail = hp->head = cp;
		cp->next = 0;
		if (cp->mask == 0)
			hp->lastExact = cp;
		return 0;
	}

	/* Otherwise, if a mask is not specified, add this to the
	 * end of all those conditions without masks.  This has
	 * the effect of checking all non-wildcard conditions first,
	 * and only if no match found checking the wildcard conditions.
	 */
	if (cp->mask == 0) {
		condition_t* former;

		/* The list is non-empty, and contains non-wildcard
		 * conditions.  Put this one after the last exact
		 * condition.
		 */
		if ((former = hp->lastExact) != 0) {
			cp->next = former->next;
			former->next = cp;

			/* If the last exact condition was also the
			 * end of the list, set the tail to this
			 * condition.
			 */
			if (former == lp)
				hp->tail = cp;
		}
		/* Otherwise, the list contains only wildcard conditions.
		 * Put this one at the front.
		 */
		else {
			hp->lastExact = cp;
			cp->next = hp->head;
			hp->head = cp;
		}
	}
	/* Otherwise, just append it to the tail */
	else {
		lp->next = cp;
		hp->tail = cp;
		cp->next = 0;
	}

	return 0;
}

/* Add a file restriction clause to the list */
addFileRestriction (ac, av)
int	ac;
char**	av;
{
	fileaccess_t* fp;
	fileaccess_t** fpp;
	unsigned int hash;
	int list;

	ac--; av++;
	if (ac != 2) {
		accessFormatError = "'restrict' command requires 2 arguments";
		return 0;
	}

	/* allocate space for copy */
	fp = (fileaccess_t*)malloc (sizeof (*fp));

	/* allocate space for filename and copy in */
	fp->fileName = malloc(strlen(av[0])+1);
	strcpy (fp->fileName, av[0]);
	ac--; av++;

	/* get list number */
	list = atoi (*av);
	if (list <= 0) {
		accessFormatError = "list argument not positive integer";
		return 0;
	}
	fp->accessList = list;

	/* compute hashcode */
	hash = (namehash(fp->fileName) & NFILEMASK);

	/* install at head of list */
	fpp = &fileHead[hash];
	fp->next = *fpp;
	*fpp  = fp;

	return 0;
}


/* primitive version of cfgets function; doesn't process continuations */
static char*
cfgets (buf, num, file, lineCount)
char*	buf;
int	num;
FILE*	file;
int	*lineCount;
{
	char* ret;

	for (;;) {
		ret = fgets(buf, num, file);
		if (ret == NULL)
			break;
		if (lineCount != (int*)0)
			*lineCount += 1;
		if (buf[0] != '#')
			break;
	}
	return ret;
}

static int
cmdValue (cmdName)
char* cmdName;
{
	register struct CMDS *cp;
	char firstChar;

	if (cmdName == (char*)0)
		return 0;

	firstChar = *cmdName;
	for (cp = configCmds; cp->cmdName != (char*)0; cp++)
		if (firstChar == *cp->cmdName && strcmp(cp->cmdName,cmdName)==0)
			return cp->cmdVal;
	return 0;
}

/* Compile a text description of an access list into
 * binary format and store it in the argument location.
 * As a side effect, set the external accessFormatError
 * to the textual description of any format error encountered.
 *
 * It returns the list for which the condition has been
 * compiled.
 *
 * Arguments are word-parsed and packed into an argument
 * vector like that used for "main".
 */
static int
compileAccessCondition (ac, av, ptr)
int	ac;
char	**av;
condition_t* ptr;
{
	int listnum;

	/* an access command must have at least 3 arguments, but
	 * not more than four.  Skip the command name itself.
	 */
	ac--, av++;
	if (ac < 3 || ac > 4) {
		accessFormatError = "illegal argument count (< 3 || > 4)";
		return 0;
	}

	/* The first argument must be numeric and must be a
	 * positive number.
	 */
	listnum = atoi (*av);
	if (listnum <= 0) {
		accessFormatError = "list number not a positive integer";
		return 0;
	}
	av++; ac--;

	/* The second argument must either be "permit", "deny", "readonly"
	 * or "readwrite".  "permit" gives old-style access.
	 */
	if (strcmp (*av, "permit") == 0)
		ptr->perm = permit;
	else if(strcmp(*av, "deny") == 0)
		ptr->perm = deny;
	else if(strcmp(*av, "readonly") == 0)
		ptr->perm = readonly;
	else if(strcmp(*av, "readwrite") == 0)
		ptr->perm = readwrite;
	else {
		accessFormatError =
			"list keyword not {permit,deny,readonly,readwrite}";
		return 0;
	}
	av++; ac--;

	/* The third argument must either be dotted decimal, or hex
	 * integer with leading "0x".
	 */
	ptr->address = inet_addr(*av);
	av++; ac--;

	/* If there is a fourth argument, it must also be either
	 * dotted decimal, or hex integer with leading "0x".
	 * If absent, mask defaults to zero.
	 */
	if (ac > 0)
		ptr->mask = inet_addr(*av);
	else
		ptr->mask = 0;

	return listnum;
}

/* Return whether the file named by the argument stat structure
 * is the same as the configuration file.
 */
isConfigFile (sb)
struct stat* sb;
{
	if (sb->st_dev == configStat.st_dev
	    && sb->st_ino == configStat.st_ino)
		return 1;
	return 0;
}

/* Get the access list which covers <fileName>.  If
 * no explict specification for this file, return the
 * default access list.  If <fileName> is fully
 * qualified (starts with '/') check to see if the
 * prefix matches the default directory.
 */
static int
getAccessList (fileName)
char* fileName;
{
	fileaccess_t* fp;
	unsigned int hash;
	char* rindex();

	if (*fileName == '/') {
		char* sep = rindex (fileName, '/');
		int count = sep - fileName;

		if (count > 0)
			count--;
		if (tftpDefaultDirectory
		    && strncmp(fileName, tftpDefaultDirectory)==0) {
			fileName = sep+1;
		}
	}

	hash = (namehash(fileName) & NFILEMASK);

	/* find access list for this file, or default
	 * access list if specified.
	 */
	for (fp = fileHead[hash]; fp != 0; fp = fp->next)
		if (strcmp(fp->fileName,fileName)==0)
			return fp->accessList;

	return defaultAccessList;
}

/* compute a hash value for a string and return it */
static unsigned int
namehash (name)
register char* name;
{
	unsigned int val;

	val = 0;
	while (*name) {
		val <<= 1;
		val += *name++;
	}
	return val;
}

/* allocate space for a string and copy it there */
static char*
newString (arg)
char* arg;
{
	char* tmp;
	if (arg == NULL)
		return NULL;

	tmp = malloc(strlen(arg)+1);
	strcpy (tmp, arg);
	return tmp;
}


/* (re)-read the tftpd config file.  We keep an open handle
 * on the file and check to see if it's been modified since
 * the last read.
 */

readConfigFile(argc, argv)
int	argc;
char**	argv;
{
	static FILE* configf;
	static unsigned long configModTime;
	char	buf[256];
	int	cargc;
	char*	cargv[32];
	char*	fileName;
	struct stat statb;

	/* If given an argument, it must be the name
	 * of a configuration file.  Otherwise use
	 * the default.
	 */
	if (tftpDebugLevel > 3) {
		syslog(LOG_DEBUG, "readConfigFile(%d,0x%08x)", argc, argv);
	}
	if (configf == 0) {
		fileName = dfltConfigFile;
		if ((argc--,argv++), argc > 0) {
			if (access(*argv, 0) == 0)
				fileName = *argv;
		}
		if ((configf = fopen (fileName, "r")) == NULL) {
			syslog (LOG_ERR,
				"Cannot open configFile '%s'; reason = %s",
				fileName, sys_errlist[errno]);
			return 0;
		}
	}
	if (tftpDebugLevel > 3) {
		syslog(LOG_DEBUG, "configFile = '%s'; FILE* = 0x%08x",
			fileName, configf);
	}

	/* See if the file has been changed since the last time we
	 * read it.
	 * @@@ This needs to use "stat" instead, in case people
	 * @@@ use editors which replace the file instead of over-write
	 * @@@ it.
	 */
	fstat(fileno(configf), &statb);
	configStat = statb;
	if (statb.st_mtime == configModTime)
		return;
	configModTime = statb.st_mtime;

	/* Read the config file */
	rewind (configf);
	resetConfig();
	configLineNumber = 0;
	while (cfgets (buf, sizeof buf, configf, &configLineNumber) != NULL) {
		char* end;

		if (end = strchr(buf, '\n'))
			*end = '\0';
		
		if (*buf == '\0')
			continue;
		cargc = split (buf, cargv, NITEMS(cargv), (char*)0);
		if (cargc == 0)
			continue;

		switch (cmdValue(cargv[0])) {
		  /* specify default directory */
		  case CMD_DEFAULT_DIR:
			if (cargc != 2)
				break;
			if (chdir(cargv[1])) {
				if (tftpDebugLevel > 0) {
					syslog (LOG_DEBUG,
						"chdir fails; '%s'",
						sys_errlist[errno]);
				}
			}
			else
				tftpDefaultDirectory = newString(cargv[1]);
			break;

		  /* specify virtual root directory */
		  case CMD_ROOT_DIR:
			if (cargc == 2) {
				if (access(cargv[1], 0) != 0)
					syslog(LOG_DEBUG, "directory '%s': %s",
						cargv[1], sys_errlist[errno]);
				else
					tftpRootDirectory = newString(cargv[1]);
			}
			break;

		  /* set debugging level */
		  case CMD_DEBUG_LEVEL:
			if (cargc != 2)
				break;
			tftpDebugLevel = atoi(cargv[1]);
			break;

		  /* create or extend an access list */
		  case CMD_ACCESS_LIST:
			addAccessList(cargc, cargv);
			break;

		  /* specify default access list */
		  case CMD_DEFAULT_ACCESS_LIST:
			if (cargc == 2)
				defaultAccessList = atoi(cargv[1]);
			break;

		  /* restrict access to a file by access list */
		  case CMD_RESTRICT:
			addFileRestriction(cargc, cargv);
			break;

		  /* set input timeout period */
		  case CMD_INPUT_WAIT:
			{ int n;
			  if (cargc != 2)
				break;
			  n = atoi (cargv[1]);
			  if (n > 0)
				maxInputWait = n;
			}
			break;
			

		  default:
			/* log illegal configuration options */
			syslog (LOG_DEBUG,
			       "Line %d: illegal configuration keyword '%s'",
				configLineNumber, cargv[0]);
			break;
			
		}
	}
	if (tftpDebugLevel > 0) {
		static char logInfo[] =
		   "init_config: default=%s; root=%s; debug=%d";
		
		syslog (LOG_DEBUG, logInfo,
			(tftpDefaultDirectory ? tftpDefaultDirectory : ""),
			(tftpRootDirectory ? tftpRootDirectory : ""),
			tftpDebugLevel);
	}
	if (tftpDebugLevel > 4) {
		accessDebugDump();
	}

	return 1;
}

/* realPath(char* orig, char* buf)
 * Deposit into <buf> the compressed pathname derived from <orig>.
 * This deletes any "." and ".." components, as well as redundant
 * separator characters.
 *
 * The re-created path always starts with a '/'.
 *
 * UNIX-specific.
 */
#define MAXCOMPONENT	256
#define MAXPATH		1024
#define SEPCHAR		'/'
#define ISCURDIR(S)	(strcmp(S,thisDir)==0)
#define ISPARENTDIR(S)	(strcmp(S,parentDir)==0)

static char	thisDir[]	= ".";
static char	parentDir[]	= "..";
static char	sepDir[]	= "/";
static char	rootDir[]	= "/";

int
realPath(orig, buf)
char*	orig;
char*	buf;
{
	char*	obuf[MAXCOMPONENT];
	char*	nbuf[MAXCOMPONENT];
	char	tmp[MAXPATH];
	int	numComponents;
	int	pathSize;
	int	num;

	/* Count the maximum number of components which could be
	 * in the source path.
	 */
	{
		register char* p;

		numComponents = 1;
		for (p = orig; *p != '\0'; p++)
			if (*p == SEPCHAR)
				numComponents++;
	}

	/* make sure the stack locals have enough space to hold
	 * everything.  If not return failure.
	 */
	if (numComponents >= MAXCOMPONENT)
		return -1;
	pathSize = strlen (obuf) + 1;
	if (pathSize >= MAXPATH)
		return -1;

	/* copy in path to temporary */
	strcpy (tmp, orig);

	/* dissect it into components */
	{
		register char* p, *q;

		num = 0;
		for (p = q = tmp; p != 0; p = q) {

			/* skip any separators */
			while (*p == SEPCHAR)
				p++;

			/* if at end of string quit */
			if  (*p == '\0')
				break;

			/* store this component away */
			obuf[num++] = p;

			/* get end of this component and replace
			 * with string terminator
			 */
			q = strchr (p, SEPCHAR);
			if (q != 0)
				*q++ = '\0';
			
		}
	}

	/* squeeze out current and parent entries */
	{
		register i, j;

		for (i=0, j=0; i < num; i++) {
			char* cur = obuf[i];
			if (ISCURDIR(cur))
				continue;
			else if (ISPARENTDIR(cur)) {
				if (j > 0)
					j--;
			}
			else
				nbuf[j++] = cur;
		}
		nbuf[j] = 0;
		num = j;
	}

	/* re-create copy of path */
	buf[0] = '\0';
	if (num == 0)
		strcpy (buf, rootDir);
	else {
		register int i;
		for (i = 0; i < num; i++) {
			strcat (buf, sepDir);
			strcat (buf, nbuf[i]);
		}
	}

	return 0;
}


/* reset configuration options to defaults */
resetConfig()
{
	register int i;
	extern char* dfltDefaultDirectory;
	extern char* dfltRootDirectory;
	extern int dfltDebugLevel;
	extern char* dfltConfigFile;

	/* free any access lists */
	if (maxAccessList) {
		for (i = 1; i <= maxAccessList; i++) {
			register condition_t *p, *q;
			for (p = accessList[i].head; p != 0; p = q) {
				q = p->next;
				free ((char*)p);
			}
		}
		maxAccessList = 0;
		bzero ((char*)accessList, maxAccessList*sizeof(accessList[0]));
	}

	/* free the file restriction lists */
	for (i = 0; i < NFILEHEAD; i++) {
		register fileaccess_t* p, *q;
		for (p = fileHead[i]; p != 0; p = q) {
			q = p->next;
			free(p->fileName);
			free((char*)p);
		}
		fileHead[i] = 0;
	}

	/* reset debug level */
	tftpDebugLevel = dfltDebugLevel;

	/* reset default directory */
	if (tftpDefaultDirectory) {
		free (tftpDefaultDirectory);
		tftpDefaultDirectory = 0;
		if (dfltDefaultDirectory)
			tftpDefaultDirectory = newString(dfltDefaultDirectory);
	}

	/* reset root directory */
	if (tftpRootDirectory) {
		free (tftpRootDirectory);
		tftpRootDirectory = 0;
		if (dfltRootDirectory)
			tftpRootDirectory = newString(dfltRootDirectory);
	}
}

/* primitive version of split function; ignores <sep> argument */
static int
split (buf, vector, count, sep)
char*	buf;
char**	vector;
int	count;
char*	sep;
{
	int nfields = 0;
	register char* bp;

	bp = buf;
	for (;;) {
		/* skip blanks */
		while (*bp == ' ' || *bp == '\t')
			bp++;

		/* if end of string, quit */
		if (*bp == '\0')
			break;

		/* put pointer to start of string in vector */
		vector[nfields++] = bp;

		/* find end of string */
		while (*bp != ' ' && *bp != '\t' && *bp != '\0')
			bp++;

		if (*bp == '\0')
			break;

		/* terminate this string and continue */
		*bp++ = '\0';
	}
	vector[nfields] = (char*)0;
	return nfields;
}

/* validate that the source system named by:
 * 'source' is permitted to access the
 * given filename according to access lists.
 *
 * Return 1 if access is permitted under old default rules
 * (i.e., no access list governs); 0 if not.
 *
 * Return 2 if more particular access is permitted by an access list.
 * This gets around the old public read/writable requirements.
 */ 
int
validateAccessToFile (filename, source, type)
char* filename;
unsigned long source; 
int type;
{
	int list;
	listhead_t* hp;
	condition_t *cp;
	enum PERM permission = deny;

	list = getAccessList(filename);

	/* No list governing this file, return OK */
	if (list == 0)
		return 1;

	/* If the named list doesn't exist, use the default */
	hp = &accessList[list];
	if (list > maxAccessList || hp->head == 0)
		if ((list = defaultAccessList) == 0)
			return 1;

	/* If that doesn't exist, return OK */
	hp = &accessList[list];
	if (list > maxAccessList || hp->head == 0)
		return 1;

	/* cycle through the conditions in this access list and
	 * check against source.  Search ends on first match.
	 */
	for (cp = hp->head; cp != 0; cp = cp->next) {
		if ((source & ~cp->mask) == cp->address) {
			permission = cp->perm;
			break;
		}
	}
	if (tftpDebugLevel > 1) {
		struct in_addr t;
		t.s_addr = source;
		syslog(LOG_DEBUG, "access to '%s' for %s is %s in list %d",
			filename, inet_ntoa(t), PERMNAME(permission),
			list);
	}

	/* Check permissions:
	 * If type is 'deny' refuse.
	 * If type is 'permit' return old default access permission
	 * If type is 'readonly' and asking for write permission, refuse.
	 * Otherwise accept.
	 */
	if (permission == deny)
		return 0;	/* DENY */
	if (permission == permit)
		return 1;	/* OK if file publically read/writable */
	if (type == WRQ && permission != readwrite)
		return 0;	/* DENY */

	/* return value of 2 means access is permitted even if
	 * the file is not publically read/writable.
	 */
	return 2;		/* OK if file is read/writable by TFTPD */
}


/* Debugging functions */


#ifndef DEBUG
accessDebugDump()
{
}
#else
#include <stdio.h>

accessDebugDump()
{
	dumpAccessLists();
	dumpFileRestrictions();
}

/* Debugging function */
static
dumpAccessLists()
{
	register int i;

	syslog (LOG_DEBUG, "default access list = %d", defaultAccessList);
	for (i = 1; i <= maxAccessList; i++) {
		if (accessList[i].head != 0) {
			syslog (LOG_DEBUG, "access list %d:", i);
			printAccessList (accessList[i].head);
		}
	}
}

static
printAccessList(list)
condition_t* list;
{
	char* permission;
	char addrbuf[32], maskbuf[32];
	struct in_addr t;

	while (list != 0) {
		permission = PERMNAME(list->perm);
		t.s_addr = list->address;
		strcpy (addrbuf, inet_ntoa(t));
		t.s_addr = list->mask;
		strcpy (maskbuf, inet_ntoa(t));
		syslog (LOG_DEBUG, " >> %s %s %s",
			permission, addrbuf, maskbuf);
		list = list->next;
	}
}

static
dumpFileRestrictions()
{
	register int i;

	syslog (LOG_DEBUG, "File Restrictions:");
	for (i = 0; i < NFILEHEAD; i++)
		if (fileHead[i] != 0)
			printFileRestriction(fileHead[i]);
}

static
printFileRestriction(list)
fileaccess_t* list;
{
	while (list != 0) {
		syslog (LOG_DEBUG, " >> file %s: list %d",
			 list->fileName, list->accessList);
		list = list->next;
	}
}
#endif
